<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * @package mod
 * @subpackage dataplus
 * @copyright 2011 The Open University
 * @license http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

/**
 * @param string $feature FEATURE_xx constant for requested feature
 * @return mixed True if module supports feature, null if doesn't know
 */
function dataplus_supports($feature) {
    switch($feature) {
        case FEATURE_GROUPS:                  return true;
        case FEATURE_GROUPINGS:               return true;
        case FEATURE_GROUPMEMBERSONLY:        return true;
        case FEATURE_MOD_INTRO:               return true;
        case FEATURE_BACKUP_MOODLE2:          return true;
        case FEATURE_COMPLETION_TRACKS_VIEWS: return true;
        case FEATURE_COMPLETION_HAS_RULES:    return true;

        default: return null;
    }
}

/**
 * add an instance to the dataplus table
 *
 * @param course $dataplus
 * @return int
 */
function dataplus_add_instance($dataplus) {
    global $DB;

    if (empty($data->assessed)) {
        $dataplus->assessed = 0;
    }

    $dataplus->timemodified = time();
    if (! $dataplus->id = $DB->insert_record('dataplus', $dataplus)) {
        return false;
    }

    return $dataplus->id;
}


/**
 * update the instance record in the dataplus table and initiate a check of the capabilities
 *
 * @param course $dataplus
 * @return mixed
 */
function dataplus_update_instance($dataplus) {
    global $DB;

    $dataplus->timemodified = time();
    $dataplus->id = $dataplus->instance;

    if (!$DB->update_record('dataplus', $dataplus)) {
        return false;
    }

    return true;
}


/**
 * delete dataplus supporting files.
 *
 * @param int $id
 * @return boolean
 */
function dataplus_delete_instance($id) {
    global $CFG;

    $cm = get_coursemodule_from_instance('dataplus', $id);
    $context = get_context_instance(CONTEXT_MODULE, $cm->id);
    $fs = get_file_storage();
    $fs->delete_area_files($context->id);

    return true;
}


/**
 * Return a small object with summary information about what a
 * user has done with a given particular instance of this module
 * Used for user activity reports.
 *
 * @param int $course Course id
 * @param int $user User id
 * @param int $mod
 * @param int $scorm The scorm id
 * @return object
 */
function dataplus_user_outline($course, $user, $mod, $dataplus) {
    global $DB;
    $sql = '';
    $actions = dataplus_get_view_actions();
    $result = new stdClass;

    foreach ($actions as $action) {
        $sql .= " (SELECT COUNT(\"action\")
                   FROM \"{log}\"
                   WHERE \"action\" = '{$action}'
                   AND \"course\" = {$course->id}
                   AND \"userid\" = {$user->id}
                   AND \"cmid\" = {$mod->id}) as {$action},";
    }

    $sql = substr($sql, 0, -1);
    $sql = "SELECT DISTINCT userid," . $sql . " FROM \"{log}\" WHERE \"userid\" = {$user->id}";

    if ($actionsresult = $DB->get_record_sql($sql)) {
        $summary = '';

        foreach ($actions as $action) {
            if ($actionsresult->$action > 0) {
                $resact = $actionsresult->$action;
                $summary .= get_string('useroutline_'.$action, 'dataplus').$resact.'<br/>';
            }
        }

        $result->info = $summary;

        $sql = "SELECT DISTINCT time
                FROM \"{log}\"
                WHERE \"course\" = {$course->id}
                AND \"userid\" = {$user->id}
                AND \"cmid\" = {$mod->id}";

        $timeresult = $DB->get_record_sql($sql, null, IGNORE_MULTIPLE);
        $result->time = $timeresult->time;
    } else {
        $result->info = get_string('useroutline_noactivity', 'dataplus');
    }

    return $result;
}


/**
 * Print a detailed representation of what a user has done with
 * a given particular instance of this module, for user activity reports.
 *
 * @param int $course
 * @param int $user
 * @param int $mod
 * @param int $scorm
 * @return boolean
 */
/*
function dataplus_user_complete($course, $user, $mod, $dataplus) {
    $sql = " (SELECT \"info\", \"time\"
              FROM \"{log}\"
              WHERE \"course\" = {$course->id}
              AND \"userid\" = {$user->id}
              AND \"cmid\" = {$mod->id}) as {$action}";

    if ($logs = $DB->get_records_sql($sql)) {
        print_simple_box_start();

        echo $user->id;

        foreach ($logs as $log) {
            echo date('d/m/Y H:i:s', $log->time).' '.$log->description.' '.'<br/>';
        }

        print_simple_box_end();
    } else {
        print_string('useroutline_noactivity', 'dataplus');
    }
}
*/


/**
 * Serves associated files
 *
 * @param object $course
 * @param object $cm
 * @param object $context
 * @param string $filearea
 * @param array $args
 * @param bool $forcedownload
 * @return mixed
 */
function dataplus_pluginfile($course, $cm, $context, $filearea, $args, $forcedownload) {
    global $DB;

    if ($context->contextlevel != CONTEXT_MODULE) {
        return false;
    }

    require_course_login($course, true, $cm);

    $postid = (int)array_shift($args);

    if (!$dataplus= $DB->get_record('dataplus', array('id'=>$cm->instance))) {
        return false;
    }

    $fs = get_file_storage();
    $relativepath = implode('/', $args);
    $fullpath = "/$context->id/mod_dataplus/$filearea/$postid/$relativepath";
    if (!$file = $fs->get_file_by_hash(sha1($fullpath)) or $file->is_directory()) {
        return false;
    }

    // Make sure groups allow this user to see this file
    /**
     * TODO - Security for group access for supporting files.
     */

    // finally send the file
    send_stored_file($file, 0, 0, true); // download MUST be forced - security!
    exit;
}


/**
 * Prints all the records uploaded by this user
 *
 * @global object
 * @param object $course
 * @param object $user
 * @param object $mod
 * @param object $data
 */
function dataplus_user_complete($course, $user, $mod, $dataplus) {
    global $DB, $CFG, $OUTPUT;
    require_once("$CFG->libdir/gradelib.php");

    $grades = grade_get_grades($course->id, 'mod', 'data', $dataplus->id, $user->id);
    if (!empty($grades->items[0]->grades)) {
        $grade = reset($grades->items[0]->grades);
        echo $OUTPUT->container(get_string('grade').': '.$grade->str_long_grade);
        if ($grade->str_feedback) {
            echo $OUTPUT->container(get_string('feedback').': '.$grade->str_feedback);
        }
    }

    echo "<p>";

    $basicconditions = array('cmid'=>$dataplus->id, 'userid'=>$user->id);
    $conditions = array_merge($basicconditions, array('action'=>'view'));

    if ($records = $DB->get_records('log', $conditions, 'time DESC')) {
        echo get_string("usercompleted_view", "dataplus", count($records)).'<br/>';
    } else {
        echo get_string("usercompleted_view", "dataplus", 0).'<br/>';
    }

    $conditions = array_merge($basicconditions, array('action'=>'insert'));

    if ($records = $DB->get_records('log', $conditions, 'time DESC')) {
        echo get_string("usercompleted_insert", "dataplus", count($records)).'<br/>';
    } else {
        echo get_string("usercompleted_insert", "dataplus", 0).'<br/>';
    }

    $conditions = array_merge($basicconditions, array('action'=>'update'));

    if ($records = $DB->get_records('log', $conditions, 'time DESC')) {
        echo get_string("usercompleted_updated", "dataplus", count($records)).'<br/>';
    } else {
        echo get_string("usercompleted_updated", "dataplus", 0).'<br/>';
    }

    $conditions = array_merge($basicconditions, array('action'=>'delete'));

    if ($records = $DB->get_records('log', $conditions, 'time DESC')) {
        echo get_string("usercompleted_deleted", "dataplus", count($records)).'<br/>';
    } else {
        echo get_string("usercompleted_deleted", "dataplus", 0).'<br/>';
    }
    echo "</p>";
}


/**
 * @return array
 */
function dataplus_get_view_actions() {
    return array('view', 'search');
}


/**
 * @return array
 */
function dataplus_get_post_actions() {
    return array('insert', 'update', 'delete');
}


/**
 * Sets the module uservisible to false if the user has not got the view capability
 * @param cm_info $cm
 */
function dataplus_cm_info_dynamic(cm_info $cm) {
    if (!has_capability('mod/dataplus:view',
            get_context_instance(CONTEXT_MODULE, $cm->id))) {
        $cm->uservisible = false;
        $cm->set_available(false);
    }
}


/**
 * Returns all other caps used in module
 */
function dataplus_get_extra_capabilities() {
    return array('moodle/site:accessallgroups');
}


/**
 * File browsing support for dataplus.
 * @param object $browser
 * @param object $areas
 * @param object $course
 * @param object $cm
 * @param object $context
 * @param string $filearea
 * @param int $itemid
 * @param string $filepath
 * @param string $filename
 * @return file_info instance Representing an actual file or folder (null if not found
 * or cannot access)
 */
function dataplus_get_file_info($browser, $areas, $course, $cm, $dpcontext, $filearea,
        $itemid, $filepath, $filename) {
    global $CFG, $DB, $dataplusfilehelper, $dataplus, $context;

    $context = get_context_instance(CONTEXT_MODULE, $cm->id);;

    if ($context->contextlevel != CONTEXT_MODULE) {
        return null;
    }

    $fileareas = array('image', 'file');
    if (!in_array($filearea, $fileareas)) {
        return null;
    }

    if (!has_capability('mod/dataplus:view', $context)) {
        return null;
    }

    require_once($CFG->dirroot . '/mod/dataplus/sqlite3_db_dataplus.php');
    require_once($CFG->dirroot . '/mod/dataplus/dataplus_file_helper.php');

    if (! $dataplus = $DB->get_record("dataplus", array("id"=>$cm->instance))) {
        print_error("Course module is incorrect");
    }
    $dataplusfilehelper = new dataplus_file_helper($dataplus->id, $context);
    $dataplusdb = new sqlite3_db_dataplus();

    $rowid = $dataplusfilehelper->get_rowid($itemid);
    $parameters[0]->name = 'id';
    $parameters[0]->value = $rowid;
    $parameters[0]->operator = 'equals';
    $row = $dataplusdb->query_dataplus_database_single(null, $parameters);

    // Make sure groups allow this user to see this file
    if (!empty($row->group_id)) {
        if (groups_get_activity_groupmode($cm, $course) == SEPARATEGROUPS) {
            // Groups are being used
            if (!groups_group_exists($row->group_id)) {
                // Can't find group
                return null;
            }
            if (!has_capability('moodle/site:accessallgroups', $context) &&
                    !groups_is_member($row->group_id)) {
                return null;
            }
        }
    }

    $fs = get_file_storage();
    $filepath = is_null($filepath) ? '/' : $filepath;
    $filename = is_null($filename) ? '.' : $filename;

    if (!($storedfile = $fs->get_file($context->id, 'mod_dataplus', $filearea, $itemid,
            $filepath, $filename))) {
        return null;
    }

    $urlbase = $CFG->wwwroot . '/pluginfile.php';
    return new file_info_stored($browser, $context, $storedfile, $urlbase, $filearea,
            $itemid, true, true, false);
}